Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Data
Imports System.Drawing
Imports System.Text
Imports System.Windows.Forms
Imports System.Net
Imports System.Net.Sockets
Imports System.Threading
Imports System.IO
Imports ICSharpCode.SharpZipLib.Checksums

Namespace FileTransfer

    Partial Public Class Form1
        Inherits Form
        Private connector As Thread
        Private client As TcpClient
        Private stream As NetworkStream
        Private ip As IPAddress
        Private connectorRunning As Boolean
        Private ReadOnly port As Integer = 21341
        Private ReadOnly STREAM_PREFIX As Integer = -2
        Private fileToSend As String

        Public Sub New()
            InitializeComponent()
        End Sub

        Private Sub Form1_Load(sender As Object, e As EventArgs)
            SetStatus(False)
        End Sub

        Private Sub btnConnect_Click(sender As Object, e As EventArgs)
            If connectorRunning Then
                log("Connector is already running.")
                Return
            End If
            Try
                ip = IPAddress.Parse(txtIP.Text)
            Catch ee As Exception
                log(ee.Message)
            End Try
            connectorRunning = True
            connector = New Thread(AddressOf startConnection)
            connector.Name = "Connector"
            connector.IsBackground = True
            connector.Start()
        End Sub

        Private Sub SetStatus(connected As Boolean)
            Dim mi As MethodInvoker = Sub()
                                          btnSend.Enabled = connected
                                          btnConnect.Enabled = Not connected
                                          lblStatus.Text = If(connected, "Connected", "Disconnected")
                                      End Sub
            Me.BeginInvoke(mi)
        End Sub

        Private Sub log(msg As String)
            Dim mi As MethodInvoker = Sub()
                                          txtLogs.AppendText(msg)
                                          txtLogs.AppendText(vbCr & vbLf)
                                          txtLogs.SelectionStart = txtLogs.Text.Length
                                          txtLogs.ScrollToCaret()
                                      End Sub
            If Me.InvokeRequired Then
                Me.BeginInvoke(mi)
            Else
                mi.Invoke()
            End If
        End Sub

        Private Sub startConnection()
            Try
                client = New TcpClient()
                client.Connect(ip, port)
                log("Connected!")
                SetStatus(True)
                stream = client.GetStream()
                Dim br As New BinaryReader(stream)
                Dim fileName As String = ""
                While connectorRunning
                    Dim prefix As Integer = br.ReadInt32()
                    If prefix > 0 Then
                        'regular message
                        Dim buffer As Byte() = br.ReadBytes(prefix)
                        fileName = Encoding.UTF8.GetString(buffer)
                        log("Receiving: " & fileName)
                    ElseIf prefix = STREAM_PREFIX Then
                        'stream message
                        Dim length As Long = br.ReadInt64()
                        'total stream length
                        Dim adler As New Adler32()
                        Using fs As New FileStream(fileName, FileMode.Create)
                            Dim buffer As Byte() = New Byte(8191) {}
                            Dim received As Long = 0
                            While received < length
                                Dim remain As Long = length - received
                                Dim count As Integer = stream.Read(buffer, 0, Math.Min(buffer.Length, CInt(remain)))
                                fs.Write(buffer, 0, count)
                                adler.Update(buffer, 0, count)
                                received += count
                            End While
                        End Using
                        Dim checkSum As Long = br.ReadInt64()
                        If checkSum <> adler.Value Then
                            Throw New Exception("Checksum does not match.")
                        End If
                        log("File received successfully.")
                    Else
                        Throw New Exception("Unexpected value: " & prefix)
                    End If
                End While
            Catch e As Exception
                log(e.Message)
                SetStatus(False)
                connectorRunning = False
            End Try
        End Sub

        Private Sub btnSend_Click(sender As Object, e As EventArgs)
            If openFileDialog1.ShowDialog(Me) = DialogResult.OK Then
                btnSend.Enabled = False
                fileToSend = openFileDialog1.FileName
                Dim sendThread As New Thread(AddressOf sendFile)
                sendThread.Name = "SendThread"
                sendThread.IsBackground = True
                sendThread.Start()
            End If
        End Sub

        Private Sub sendFile()
            Try
                Dim adler As New Adler32()
                log("Sending: " & fileToSend)
                Using fs As New FileStream(fileToSend, FileMode.Open)
                    Dim name As Byte() = Encoding.UTF8.GetBytes(Path.GetFileName(fileToSend))
                    Dim bw As New BinaryWriter(stream)
                    'send the first message which is the file name
                    bw.Write(name.Length)
                    bw.Write(name)
                    'send the stream message
                    bw.Write(STREAM_PREFIX)
                    Dim size As Long = New FileInfo(fileToSend).Length
                    bw.Write(size)
                    Dim written As Long = 0
                    Dim buffer As Byte() = New Byte(8191) {}
                    While written < size
                        Dim count As Integer = fs.Read(buffer, 0, Math.Min(CInt(size - written), buffer.Length))
                        stream.Write(buffer, 0, count)
                        adler.Update(buffer, 0, count)
                        written += count
                    End While
                    bw.Write(adler.Value)
                    log("File sent successfully.")
                End Using
            Catch e As Exception
                log(e.Message)
            End Try
            Dim mi As MethodInvoker = Sub() btnSend.Enabled = Not btnConnect.Enabled
            Me.BeginInvoke(mi)
        End Sub
    End Class

End Namespace
